/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <defs.h>
#include <packet.h>
#include <eventhandler.h>
#include <squeue.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <rng.h>
#include <trgen.h>
#include <trdump.h>
#include <intransit.h>
#include <codeset.h>
#include <debug.h>
#include <txsched.h>

extern Scheduler * sched;
extern InTransit * it;
extern char inputfile[20];
extern Codeset *codeset;
extern TrDump *trdump;
extern TxSched * txsched;


TrGen::TrGen(int t ):EventHandler()
{
	fin = fopen(inputfile, "r");
	if (fin == NULL)
	{
		dtprintf("trgen.cc: TrGen::TrGen(): Could not open input file : %s for reading\n", inputfile);
	}
	rng = new RNG();
	lastarrival = 0;
	done = 0;
	fout = fopen("input.trc", "w");	
	debug = 0;

	traffic = t;
}

TrGen::~TrGen()
{
	if(fin != NULL)
		fclose(fin);
}


//The data acquisition function of the TrGen: This function reads the 
//trace file, constructs a packet and processes it.
void TrGen::read()
{
	dtprintf("trgen.cc: TrGen::read(): Entered function\n");

	static int numpkts =0;
	static int id = 0;	

	double arrivaltime;
	int pktid;
	int srcaddr1 = 0;
	int srcaddr2 = 0;
	int srcaddr3 = 0;
	int srcaddr4 = 0;
	int srcport;
	int dstaddr1 = 0;
	int dstaddr2 = 0;
	int dstaddr3 = 0;
	int dstaddr4 = 0;
	int dstport;
	int size;

	double x, y, z;

//Read event from trace file
	if(traffic == TRAFFIC_FILE)
	{
		if (fscanf(fin,  "%lf %lf %lf %lf %d %d.%d.%d.%d.%d %d.%d.%d.%d.%d %d\n", &arrivaltime, &x, &y, &z,  &size,  &srcaddr1,&srcaddr2, &srcaddr3, &srcaddr4, &srcport,  &dstaddr1, &dstaddr2, &dstaddr3, &dstaddr4, &dstport, &pktid) == EOF)
		{
			//sched->halt = 1;
			dtprintf("trgen.cc: TrGen::read(): No more packets.\n");
			dprintf("\n");
			done =1;	
			totalpkts = numpkts;
			return;
		}
		if(pktid == 10000)
		{
			totalpkts = numpkts;
			return;
		}
	}
	else
	{
		arrivaltime = genarrivaltime();
		size = (int) genpacketsize();
		srcaddr4 = dstaddr4 = srcport = dstport = 1;
		x = 0;
		y = 0;
		z = 0;
		x = gendistance();
		pktid = id;
		id++;
		if(id == totalpkts)
		{
			done =1;	
			return;
		}

		//Dump the traffic file for future reference/stats calcs
		fprintf(fout,  "%.12f %.6f %.6f %.6f %d %d.%d %d.%d %d\n", arrivaltime, x, y, z,  size,  srcaddr4, srcport, dstaddr4, dstport, pktid);
	}


//Call ipport's processing function.

	Packet *packet = new Packet();
	packet->id = pktid;
	packet->srcaddr = (int)(srcaddr4 + srcaddr3* pow(2, 8) + srcaddr2* pow(2, 16) + srcaddr1* pow(2, 24));
	packet->dstaddr = (int)(dstaddr4 + srcaddr3* pow(2, 8) + dstaddr3* pow(2, 16) + dstaddr1 *pow(2, 24));
	packet->srcport = srcport;
	packet->dstport = dstport;
	packet->size = size;
	dtprintf("trgen.cc: TrGen::read(): Packetid: %d Src addr: %d Dst addr: %d Size: %d\n", pktid, packet->srcaddr, packet->dstaddr, size);
	dtprintf("trgen.cc: TrGen::read(): Distance: %.6f\n", x);
	dtprintf("trgen.cc: TrGen::read(): Current time: %.12f\n", sched->currenttime);
	dtprintf("trgen.cc: TrGen::read(): Timestamp: %.12f\n", arrivaltime);


	if(size <= 0)
	{
		printf("Error in trace file: Size of packet <= 0: %d \n", pktid);
		exit(-1);
	}

//TBD: Fix format of timestamp (double to long int)
	packet->arrivaltime = arrivaltime;
	packet->depx = x;
	packet->depy = y;
	packet->depz = z;
	packet->status = PACKET_OK;	
	packet->offset = 0;

	int chosen = 0;

//When using traffic from file, always assign the same codeword to a dst addr.
	if(traffic == TRAFFIC_FILE)
		chosen = packet->dstaddr % codeset->numcodewords;
	else
		chosen = (int) ((random()/(pow(2, 31) - 1))*(codeset->numcodewords));

	packet->codeword = chosen;

	dtprintf("trgen.cc: TrGen::read(): Codeword index: %d\n", chosen);
	packet->A = codeset->A;
	packet->N = codeset->N;
	packet->w = codeset->w;
	dtprintf("trgen.cc: TrGen::read()::Code :");	
	for (int count1 =0; count1 < packet->A; count1++)
	{
		for (int count =packet->N - 1; count >=0 ; count--)
		{
			dprintf("%d ", (codeset->set[(packet->codeword) * (packet->A*packet->N) + packet->N*count1 + count ] ));
		}
	}
	dprintf("\n");


	packet->chiprate = codeset->chiprate;
	packet->bitrate = packet->chiprate/codeset->N;

	double delay  = arrivaltime - sched->currenttime;
	dtprintf("trgen.cc: TrGen::read(): Scheduling packet for %.12f\n", delay);
	dprintf("\n");

	sched->scheduleevent(this, packet, delay );

	numpkts++;
}

//The event handling function of the TrGen. It dequeues a packet and
//sends it to the next component.
void TrGen::handleevent(Packet *p)
{
	dtprintf("trgen.cc: TrGen::handleevent(): Entered function\n");
	

	//Pass it on to the next component in the path.
	dtprintf("trgen.cc: TrGen::handleevent(): Packet id: %d\n", p->id);
	dtprintf("trgen.cc: TrGen::handleevent(): Packet processing complete, passing it on to the next component.\n");
	dprintf("\n");


	txsched->process(p);
	read();
}

void TrGen::dumpstats()
{
	printf("trgen.cc: TrGen::dumpstats():\n");
}

double TrGen::genarrivaltime()
{
	double ia = 0;
	switch(traffic)
	{
		case TRAFFIC_MM:	
		case TRAFFIC_MTM:
			ia = rng->exponential(avginterarrival);
			break;		
		case TRAFFIC_DD:
			ia = avginterarrival;
			break;
		case TRAFFIC_UU:
			ia = rng->uniform(2 * avginterarrival);
			break;
	}

//Round off to the nearest chip boundary 
	ia = ia * codeset->chiprate; 	
	int a = (int)ia;
	ia = (double)a/(codeset->chiprate);
	dtprintf("trgen.cc: TrGen::handleevent(): Departure: %.12f\n", ia);

	double newarrival = lastarrival + ia;	
	lastarrival = newarrival;
	

	
	return(newarrival);		
}

double TrGen::genpacketsize()
{
	double size =0;		
	switch(traffic)
	{
		case TRAFFIC_MM:	
			size = rng->exponential(avgpktsize);
			break;		
		case TRAFFIC_DD:
			size = avgpktsize;
			break;
		case TRAFFIC_UU:
			size = rng->uniform(2 *avgpktsize);
			break;
		case TRAFFIC_MTM:

			double p = (random()/(pow(2, 31) - 1));
			if(p < .7)
				size =40;
			else if((p < .8) && (p > .7))
				size = 576;
			else
				size = 1500;
			break;
	}
	dtprintf("trgen.cc: TrGen::handleevent(): Packet size: %.6f\n", size );
	if (size < 1.0)
		size = 1;
	return(size);

}


double TrGen::gendistance()
{

	double dist = rng->uniform(avgdistance);
	return(dist);

}
